/* ast.java
 * Author: John Lenz <jelenz@wisc.edu>
 * (C) Copyright 2003 John Lenz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.io.*;
import java.util.*;

// **********************************************************************
// The ASTnode class defines the nodes of the abstract-syntax tree that
// represents a Toy program.
//
// Internal nodes of the tree contain pointers to children, organized
// either in a linked list (for nodes that may have a variable number
// of children) or as a fixed set of fields.
//
// The nodes for literals and ids contain line and character number
// information; for string literals and identifiers, they also contain a
// string; for integer literals, they also contain an integer value.
//
// Here are all the different kinds of AST nodes and what kinds of children
// they have.  All of these kinds of AST nodes are subclasses of "ASTnode".
// Indentation indicates further subclassing:
//
//     Subclass            Kids
//     --------            ----
//     ProgramNode         DeclListNode
//     DeclListNode        linked list of DeclNode
//     StructDeclListNode  linked list of VarDeclNode
//     DeclNode:
//       VarDeclNode       TypeNode, IdNode, int
//       FnDeclNode        TypeNode, IdNode, FormalsListNode, FnBodyNode
//       FnPreDeclNode     TypeNode, IdNode, FormalsListNode
//       FormalDeclNode    TypeNode, IdNode
//       StructDeclNode    IdNode, DeclListNode
//
//     FormalsListNode     linked list of FormalDeclNode
//     FnBodyNode          DeclListNode, StmtListNode
//     StmtListNode        linked list of StmtNode
//     ExpListNode         linked list of ExpNode
//
//     TypeNode:
//       IntNode           -- none --
//       BoolNode          -- none --
//       VoidNode          -- none --
//       StructTypeNode    IdNode
//
//     StmtNode:
//       ReadStmtNode        ExpNode
//       WriteStmtNode       ExpNode
//       AssignStmtNode      ExpNode, ExpNode
//       IfStmtNode          ExpNode, DeclListNode, StmtListNode
//       IfElseStmtNode      ExpNode, DeclListNode, StmtListNode,
//                                    DeclListNode, StmtListNode
//       WhileStmtNode       ExpNode, DeclListNode, StmtListNode
//       ForStmtNode         ExpNode, DeclListNode, StmtListNode
//       CallStmtNode        CallExpNode
//       ReturnStmtNode      ExpNode
//
//     ExpNode:
//       IntLitNode          -- none --
//       StrLitNode          -- none --
//       TrueNode            -- none --
//       FalseNode           -- none --
//       IdNode              -- none --
//       ArrayExpNode        IdNode, ExpNode
//       CallExpNode         IdNode, ExpListNode
//       StructRefNode       ExpNode, IdNode
//       UnaryExpNode        ExpNode
//         UnaryMinusNode
//         NotNode
//         DeRefNode
//         AddrOfNode
//       BinaryExpNode       ExpNode ExpNode
//         PlusNode     
//         MinusNode
//         TimesNode
//         DivideNode
//         ModuloNode
//         AndNode
//         OrNode
//         EqualsNode
//         NotEqualsNode
//         LessNode
//         GreaterNode
//         LessEqNode
//         GreaterEqNode
//
// Here are the different kinds of AST nodes again, organized according to
// whether they are leaves, internal nodes with linked lists of kids, or
// internal nodes with a fixed number of kids:
//
// (1) Leaf nodes:
//        IntNode,   BoolNode,  VoidNode,  IntLitNode,  StrLitNode,
//        TrueNode,  FalseNode, IdNode
//
// (2) Internal nodes with (possibly empty) linked lists of children:
//        DeclListNode, FormalsListNode, StmtListNode, ExpListNode
//
// (3) Internal nodes with fixed numbers of kids:
//        ProgramNode,    VarDeclNode,    FnDeclNode,     FormalDeclNode,
//        FnBodyNode,     TypeNode,       ReadStmtNode,   WriteStmtNode
//        AssignStmtNode, IfStmtNode,     IfElseStmtNode, WhileStmtNode,
//        CallStmtNode,   ReturnStmtNode, ArrayExpNode,   CallExpNode,
//        UnaryExpNode,   BinaryExpNode,  UnaryMinusNode, NotNode,
//        PlusNode,       MinusNode,      TimesNode,      DivideNode,
//        AndNode,        OrNode,         EqualsNode,     NotEqualsNode,
//        LessNode,       GreaterNode,    LessEqNode,     GreaterEqNode
//
// **********************************************************************

// **********************************************************************
// ASTnode class (base class for all other kinds of nodes)
// **********************************************************************
abstract class ASTnode { 
  // every subclass must provide an unparse operation
  abstract public void unparse(PrintWriter p, int indent);
  
  // this method can be used by the unparse methods to do indenting
  protected void doIndent(PrintWriter p, int indent) {
    for (int k=0; k<indent; k++) p.print(" ");
  }
}

// **********************************************************************
// ProgramNode,  DeclListNode, FormalsListNode, FnBodyNode,
// StmtListNode, ExpListNode
// **********************************************************************
class ProgramNode extends ASTnode {
  public ProgramNode(DeclListNode L) {
    myDeclList = L;
  }

  public void unparse(PrintWriter p, int indent) {
    myDeclList.unparse(p, indent);
  }

  public void updateNames(SymTab table) {
    LinkedList l = new LinkedList();
    l.addLast(Type.Void.addPointer(1));
    l.addLast(Type.Int);
    Sym s = new Sym("sbrk", l);
    try {
      table.insert(s);
    } catch (DuplicateException e) {
      System.err.println("Unexpected DuplicateException in ProgramNode.");
      System.exit(-1);
    } catch (EmptySymTabException e) {
      System.err.println("Unexpected EmptySymTabException in ProgramNode.");
      System.exit(-1);
    }
    
    myDeclList.updateNames(table, -1);
    s = table.globalLookup("main");
    if (s == null) {
      Errors.fatal(0, 0, "No main function");
    }
  }
  
  public void codeGen() {
    Codegen.generate(".text");
    Codegen.genLabel("_sbrk");
    if (Codegen.Wisc) {
      Codegen.genPop(Codegen.T0);
      Codegen.generate("la", Codegen.T1, "heap");
      Codegen.generateIndexed("lw", Codegen.V0, Codegen.T1, 0);
      Codegen.generate(Codegen.ADD, Codegen.T0, Codegen.V0, Codegen.T0);
      Codegen.generateIndexed("sw", Codegen.T0, Codegen.T1, 0);
    } else {
      Codegen.genPop(Codegen.A0);
      Codegen.generate("li", Codegen.V0, "9");
      Codegen.generateWithComment("syscall", "sbrk");
    }
    Codegen.generate("jr", Codegen.RA);

    myDeclList.codeGen();

    if (Codegen.Wisc) {
      Codegen.generate(".data");
      if (!Codegen.Wisc) Codegen.generate(".align 2");
      Codegen.generateLabeled("heap", ".space " + Codegen.wordSize, "", "");
    }
  }

  public void checkTypes() {
    myDeclList.checkTypes();
  }

  // 1 kid
  private DeclListNode myDeclList;
}

class DeclListNode extends ASTnode {
  public DeclListNode(LinkedList S) {
    myDecls = S;
  }

  public void unparse(PrintWriter p, int indent) {
    Iterator it = myDecls.iterator();
    try {
        while (it.hasNext()) {
          ((DeclNode)it.next()).unparse(p, indent);
        }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in DeclListNode.unparse");
        System.exit(-1);
    }
  }

  public int updateNames(SymTab table, int off) {
    Iterator it = myDecls.iterator();
    try {
      while (it.hasNext()) {
        off = ((DeclNode)it.next()).updateNames(table, off, false);
      }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in DeclListNode.updateNames");
        System.exit(-1);
    }
    return off;
  }

  public void checkTypes() {
    Iterator it = myDecls.iterator();
    try {
      while (it.hasNext()) {
        ((DeclNode)it.next()).checkTypes();
      }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in DeclListNode.checkTypes");
        System.exit(-1);
    }
  }

  public void codeGen() {
    Iterator it = myDecls.iterator();
    try {
      while (it.hasNext()) {
        ((DeclNode)it.next()).codeGen();
      }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in DeclListNode.codeGen");
        System.exit(-1);
    }
  }

  // linked list of kids (DeclNodes)
  protected LinkedList myDecls;
}

class StructDeclListNode extends DeclListNode {
  public StructDeclListNode(LinkedList S) {
    super(S);
  }

  public void setStructName(String s) {
    Iterator it = myDecls.iterator();
    try {
      while (it.hasNext()) {
        ((VarDeclNode)it.next()).setStructName(s);
      }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in StructDeclListNode.setStructName");
        System.exit(-1);
    }
  }

  public int updateNames(SymTab table, int off) {
    Iterator it = myDecls.iterator();
    try {
      while (it.hasNext()) {
        off = ((DeclNode)it.next()).updateNames(table, off, true);
      }
    } catch (NoSuchElementException ex) {
        System.err.println("unexpected NoSuchElementException in StructDeclListNode.updateNames");
        System.exit(-1);
    }
    return off;
  }
}
  

class FormalsListNode extends ASTnode {
  public FormalsListNode(LinkedList S) {
    myFormals = S;
  }

  public void unparse(PrintWriter p, int indent) {
    Iterator it = myFormals.iterator();
    try {
      if (it.hasNext())
        ((FormalDeclNode)it.next()).unparse(p, 0);
      while (it.hasNext()) {
        p.print(", ");
        ((FormalDeclNode)it.next()).unparse(p, 0);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in FormalsListNode.unparse");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void addTypes(LinkedList list, SymTab table) {
    Iterator it = myFormals.iterator();
    try {
      while (it.hasNext()) {
        ((FormalDeclNode)it.next()).addTypes(list, table);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in FormalsListNode.addTypes");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public int updateNames(SymTab table, int off) {
    Iterator it = myFormals.iterator();
    try {
      while (it.hasNext()) {
        off = ((FormalDeclNode)it.next()).updateNames(table, off, false);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in FormalsListNode.updateNames");
      ex.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  // linked list of kids (FormalDeclNodes)
  private LinkedList myFormals;
}

class FnBodyNode extends ASTnode {
  public FnBodyNode(DeclListNode declList, StmtListNode stmtList) {
    myDeclList = declList;
    myStmtList = stmtList;
  }

  public void unparse(PrintWriter p, int indent) {
    myDeclList.unparse(p, indent);
    myStmtList.unparse(p, indent);
  }

  public int updateNames(SymTab table, int off) {
    off = myDeclList.updateNames(table, off);
    return myStmtList.updateNames(table, off);
  }

  /// ret is the return type of the enclosing function.
  public void checkTypes(Type ret) {
    myStmtList.checkTypes(ret);
  }

  public void codeGen(String retLabel) {
    myStmtList.codeGen(retLabel);
  }

  // 2 kids
  private DeclListNode myDeclList;
  private StmtListNode myStmtList;
}

class StmtListNode extends ASTnode {
  public StmtListNode(LinkedList S) {
    myStmts = S;
  }

  public void unparse(PrintWriter p, int indent) {
    Iterator it = myStmts.iterator();
    try {
      while (it.hasNext()) {
        ((StmtNode)it.next()).unparse(p, indent);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in StmtListNode.unparse");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public int updateNames(SymTab table, int off) {
    Iterator it = myStmts.iterator();
    try {
      while (it.hasNext()) {
        off = ((StmtNode)it.next()).updateNames(table, off);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in StmtListNode.updateNames");
      ex.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  /// ret is the return type of the enclosing function.
  public void checkTypes(Type ret) {
    Iterator it = myStmts.iterator();
    try {
      while (it.hasNext()) {
        ((StmtNode)it.next()).checkTypes(ret);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in StmtListNode.checkTypes");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void codeGen(String retLabel) {
    Iterator it = myStmts.iterator();
    try {
      while (it.hasNext()) {
        ((StmtNode)it.next()).codeGen(retLabel);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in StmtListNode.codeGen");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  // linked list of kids (StmtNodes)
  private LinkedList myStmts;
}

class ExpListNode extends ASTnode {
  public ExpListNode(LinkedList S) {
    myExps = S;
  }

  public void unparse(PrintWriter p, int indent) {
    Iterator it = myExps.iterator();
    try {
      if (it.hasNext())
        ((ExpNode)it.next()).unparse(p, 0);
      while (it.hasNext()) {
        p.print(", ");
        ((ExpNode)it.next()).unparse(p, 0);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in ExpListNode.unparse");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void updateNames(SymTab table) {
    Iterator it = myExps.iterator();
    try {
      while (it.hasNext()) {
        ((ExpNode)it.next()).updateNames(table);
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in ExpListNode.updateNames");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void checkTypes(Iterator paramIt) {
    Iterator expIt = myExps.iterator();
    try {
      while (expIt.hasNext()) {
        ExpNode n = (ExpNode)expIt.next();
        Type nodet = n.checkTypes();
        Type paramt = (Type) paramIt.next();

        if (! nodet.equals(Type.Error) && ! nodet.equals(paramt)) {
          Errors.fatal(n.getLine(), n.getChar(), "Type of actual does not match type of formal");
        }
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in ExpListNode.checkTypes");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public void codeGen() {
    Iterator it = myExps.iterator();
    try {
      while (it.hasNext()) {
        ((ExpNode)it.next()).codeGen();
      }
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in ExpListNode.codeGen");
      ex.printStackTrace();
      System.exit(-1);
    }
  }

  public int numExprs() { return myExps.size(); }

  // linked list of kids (ExpNodes)
  private LinkedList myExps;
}

// **********************************************************************
// DeclNode and its subclasses
// **********************************************************************
abstract class DeclNode extends ASTnode {
  public void checkTypes() {}
  abstract public int updateNames(SymTab sym, int off, boolean structure);
  abstract public void codeGen();
}

class VarDeclNode extends DeclNode {
  public VarDeclNode(TypeNode type, IdNode id, int size, int ptrs) {
    myType = type;
    myId = id;
    mySize = size;
    numPointers = ptrs;
    structName = "";
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    myType.unparse(p, 0);
    p.print(" ");
    myId.unparse(p, 0);
    if (mySize != NOT_ARRAY) {
        p.print("[" + mySize + "]");
    }
    p.println(";");
  }

  /// This function must deal with every possible combination
  /// of pointers, structures, and arrays and calculate the correct
  /// offsets for all of them
  public int updateNames(SymTab table, int off, boolean structure) {
    int thisOff = off;
    if (myType.myType(mySize).equals(Type.Void)) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Non-function declared void");
    } else {
      try {
        Type thisType = null;
        if (myType.myType(0).equals(Type.Struct)) {
          Sym s = table.globalLookup(myType.name());
          if (s == null) {
            Errors.fatal(myId.getLine(), myId.getChar(), "Structure not yet declared");
          } else {
            if (mySize == NOT_ARRAY) {
              if (numPointers == 0) {
                // is a structure
                mySize = s.size();
                if (s.size() < 0)
                  Errors.fatal(myId.getLine(), myId.getChar(), "Structure not yet declared");
                if (off >= 0) thisOff += mySize;
              } else {
                // is a pointer to a struct
                if (off >= 0) thisOff += Codegen.wordSize;
              } 
              thisType = Type.Struct(myType.name()).addPointer(numPointers);
            } else {
              if (numPointers == 0) {
                // an array of structures
                mySize *= s.size();
                if (s.size() < 0) 
                  Errors.fatal(myId.getLine(), myId.getChar(), "Structure not yet declared");
                if (off >= 0) thisOff += mySize;
              } else {
                // an array of pointers to structures
                if (off >= 0) thisOff += Codegen.wordSize * mySize;
              }
              thisType = Type.Struct(myType.name()).addPointer(numPointers + 1);
            }
          }
        } else {
          thisType = myType.myType(mySize, numPointers);
          // set the offset for the current decl to be the lower address if decl is an array.
          if (off >= 0) {
            if (mySize == NOT_ARRAY) {
              thisOff += Codegen.wordSize;
            } else {
              thisOff += Codegen.wordSize * mySize;
            }
          }
        }
        if (structure && mySize != NOT_ARRAY)
          myId.sym = new Sym(structName + myId.getName(), thisType, mySize, off);
        else
          myId.sym = new Sym(structName + myId.getName(), thisType, mySize, thisOff - Codegen.wordSize);

        table.insert(myId.sym);
      } catch (DuplicateException e) {
        myId.sym = null;
        Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
      } catch (EmptySymTabException e) {
        System.err.println("unexpected EmptySymTabException in VarDeclNode.updateNames");
        e.printStackTrace();
        System.exit(-1);
      }
    }
    return thisOff;
  }

  public void codeGen() {
    Codegen.generate(".data");
    if (!Codegen.Wisc) Codegen.generate(".align 2");
    if (myId.sym.offset() < 0) {
      if (mySize == NOT_ARRAY)
        Codegen.generateLabeled("_" + myId.sym.name(), ".space " + Codegen.wordSize, "", "");
      else
        Codegen.generateLabeled("_" + myId.sym.name(), ".space " + Codegen.wordSize * mySize, "", "");
    }
  }

  public void setStructName(String s) {
    structName = s + ".";
  }

  // 3 kids
  private TypeNode myType;
  private IdNode myId;
  private int mySize;  // use value NOT_ARRAY if this is not an array type
  private int numPointers;
  private String structName;

  public static int NOT_ARRAY = -1;
}

class FnDeclNode extends DeclNode {
  public FnDeclNode(TypeNode type,
                    int numPtr,
                    IdNode id,
                    FormalsListNode formalList,
                    FnBodyNode body) {
    myType = type;
    myId = id;
    myFormalsList = formalList;
    myBody = body;
    numPointers = numPtr;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    myType.unparse(p, 0);
    p.print(" ");
    myId.unparse(p, 0);
    p.print("(");
    myFormalsList.unparse(p, 0);
    p.println(") {");
    myBody.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
  }

  public int updateNames(SymTab table, int off, boolean structure) {
    LinkedList l = new LinkedList();
    l.addLast(myType.myType(numPointers));
    myFormalsList.addTypes(l, table);

    // Check if function is pre-declared
    myId.sym = table.globalLookup(myId.getName());
    if (myId.sym == null) {
      myId.sym = new Sym(myId.getName(), l);
      try {
        table.insert(myId.sym);
      } catch (DuplicateException e) {
        Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
        myId.sym = null;
      } catch (EmptySymTabException e) {
        System.err.println("unexpected EmptySymTabException in FnDeclNode.updateNames");
        e.printStackTrace();
        System.exit(-1);
      }
    } else { // function is pre-declared so check params match
      if (myId.sym.size() == -1) { // already declared
        Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
        myId.sym = null;
      } else {
        myId.sym.setSize(-1);
        Iterator i = l.iterator();
        Iterator j = myId.sym.params().iterator();
        boolean error = false;
        while (i.hasNext() && j.hasNext() && !error) {
          if (!((Type)i.next()).equals((Type)j.next())) {
            Errors.fatal(myId.getLine(), myId.getChar(), "Formal list does not match pre-decleration");
            error = true;
            myId.sym = null;
          }
        }
        if (!error && (j.hasNext() || i.hasNext())) {
          Errors.fatal(myId.getLine(), myId.getChar(), "Formal list does not match pre-decleration");
          myId.sym = null;
        }
      }
    }
      
    table.addHashtab();
      returnAddrOffset = myFormalsList.updateNames(table, 0);
      initSPOffset = myBody.updateNames(table, returnAddrOffset + Codegen.wordSize * 2);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in FnDeclNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void checkTypes() {
    if (myType.myType(numPointers).equals(Type.Struct)) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Returning a structure from a function");
    }
    myBody.checkTypes(myType.myType(numPointers));
  }

  public void codeGen() {
    Codegen.generate(".text");
    if (myId.sym.name().equals("main")) {
      if (Codegen.Wisc) {
        Codegen.genLabel("main");
        Codegen.generate("li", Codegen.SP, "0xFFFF");
        Codegen.generate("la", Codegen.T0, "heap");
        Codegen.generate(Codegen.ADD, Codegen.T1, Codegen.T0, "1");
        Codegen.generateIndexed("sw", Codegen.T1, Codegen.T0, 0);
      } else {
        Codegen.generate(".globl main");
        Codegen.genLabel("main");
      }
    } else {
      Codegen.genLabel("_" + myId.sym.name());
    }
    Codegen.genPush(Codegen.RA);
    Codegen.genPush(Codegen.FP);
    Codegen.generate(Codegen.ADD, Codegen.FP, Codegen.SP, returnAddrOffset + Codegen.wordSize * 2);
    if (Codegen.Wisc && (initSPOffset < -15 || initSPOffset > 16)) {
      Codegen.generate("li", Codegen.T0, initSPOffset);
      Codegen.generate(Codegen.SUB, Codegen.SP, Codegen.FP, Codegen.T0);
    } else {
      Codegen.generate(Codegen.SUB, Codegen.SP, Codegen.FP, initSPOffset);
    }
    String end = Codegen.nextLabel();
    myBody.codeGen(end);
    Codegen.genLabel(end);
    if (Codegen.Wisc && myId.sym.name().equals("main")) {
      Codegen.generate("halt");
    } else {
      Codegen.generateIndexed("lw", Codegen.RA, Codegen.FP, -returnAddrOffset);
      Codegen.generate("move", Codegen.T0, Codegen.FP);
      Codegen.generateIndexed("lw", Codegen.FP, Codegen.FP, -returnAddrOffset - Codegen.wordSize);
      Codegen.generate("move", Codegen.SP, Codegen.T0);
      Codegen.generate("jr", Codegen.RA);
    }
  }

  // 4 kids
  private TypeNode myType;
  private int numPointers;
  private IdNode myId;
  private FormalsListNode myFormalsList;
  private FnBodyNode myBody;
  private int returnAddrOffset;
  private int initSPOffset;
}

class FnPreDeclNode extends DeclNode {
  public FnPreDeclNode(TypeNode type,
                    int numPtr,
                    IdNode id,
                    FormalsListNode formalList) {
    myType = type;
    myId = id;
    myFormalsList = formalList;
    numPointers = numPtr;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    myType.unparse(p, 0);
    p.print(" ");
    myId.unparse(p, 0);
    p.print("(");
    myFormalsList.unparse(p, 0);
    p.println(");");
  }

  public int updateNames(SymTab table, int off, boolean structure) {
    LinkedList l = new LinkedList();
    l.addLast(myType.myType(numPointers));
    myFormalsList.addTypes(l, table);
    myId.sym = new Sym(myId.getName(), l);
    myId.sym.setSize(-2);
    try {
      table.insert(myId.sym);
    } catch (DuplicateException e) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
      myId.sym = null;
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in FnDeclNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void codeGen() {

  }

  // 4 kids
  private TypeNode myType;
  private int numPointers;
  private IdNode myId;
  private FormalsListNode myFormalsList;
}

class FormalDeclNode extends DeclNode {
  public FormalDeclNode(TypeNode type, IdNode id, int ptrs) {
    myType = type;
    myId = id;
    numPointers = ptrs;
  }

  public void unparse(PrintWriter p, int indent) {
    myType.unparse(p, 0);
    p.print(" ");
    myId.unparse(p, 0);
  }

  public void addTypes(LinkedList list, SymTab table) {
    Type t = null;
    
    if (myType.myType(0).equals(Type.Struct)) {
      Sym s = table.globalLookup(myType.name());
      if (s != null)
        t = Type.Struct(myType.name()).addPointer(numPointers);
    } else {
      t = myType.myType(numPointers);
    }
    list.addLast(t);
  }

  public int updateNames(SymTab table, int off, boolean structure) {
    try {
      Type thisType = null;
      if (myType.myType(0).equals(Type.Struct)) {
        if (numPointers > 0) {
          Sym s = table.globalLookup(myType.name());
          if (s == null) {
            Errors.fatal(myId.getLine(), myId.getChar(), "Structure not yet declared");
          } else {
            thisType = Type.Struct(myType.name()).addPointer(numPointers);
          }
        } else {
          Errors.fatal(myId.getLine(), myId.getChar(), "Passing structure by reference is not allowed");
        }
      } else {
        thisType = myType.myType(numPointers);
      }
      myId.sym = new Sym(myId.getName(), thisType, off);
      if (off >= 0) off += Codegen.wordSize;
      table.insert(myId.sym);
    } catch (DuplicateException e) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
      myId.sym = null;
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in FormalDeclNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void codeGen() {}

  // 2 kids
  private TypeNode myType;
  private IdNode myId;
  private int numPointers;
}

class StructDeclNode extends DeclNode {
  public StructDeclNode(IdNode name, StructDeclListNode types) {
    myId = name;
    myDecls = types;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("struct ");
    myId.unparse(p, 0);
    p.println(" {");
    myDecls.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("};");
  }

  public int updateNames(SymTab table, int off, boolean structure) {

    // check if struct is pre-declared
    myId.sym = table.globalLookup(myId.getName());
    if (myId.sym == null) {
      try {
        myId.sym = new Sym(myId.getName(), Type.Struct(myId.getName()), -2, 0);
        table.insert(myId.sym);
      } catch (DuplicateException e) {
        Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
        myId.sym = null;
      } catch (EmptySymTabException e) {
        System.err.println("unexpected EmptySymTabException in FormalDeclNode.updateNames");
        e.printStackTrace();
        System.exit(-1);
      }
    } else {
      if (!myId.sym.type().equals(Type.Struct) ||
          (myId.sym.type().equals(Type.Struct) && myId.sym.size() != -2)) {
        Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
        myId.sym = null;
      }
    }
    myDecls.setStructName(myId.getName());
    size = myDecls.updateNames(table, 0);
    if (myId.sym != null) myId.sym.setSize(size);
    return off;
  }

  public void codeGen() {

  }

  // 3 kids
  private IdNode myId;
  private StructDeclListNode myDecls;
  private int size;
}

class StructPreDeclNode extends DeclNode {
  public StructPreDeclNode(IdNode name)  {
    myId = name;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("struct ");
    myId.unparse(p, 0);
    p.println(";");
  }

  public int updateNames(SymTab table, int off, boolean structure) {
    try {
      myId.sym = new Sym(myId.getName(), Type.Struct(myId.getName()), -2, 0);
      table.insert(myId.sym);
    } catch (DuplicateException e) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Multiply declared identifier");
      myId.sym = null;
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in FormalDeclNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void codeGen() {

  }

  private IdNode myId;
}

// **********************************************************************
// TypeNode and its Subclasses
// **********************************************************************
abstract class TypeNode extends ASTnode {
  abstract public Type myType(int array, int num);
  public Type myType(int num) { return myType(-1, num); }
  public void unparse(PrintWriter p, int indent) { p.print(name()); }
  abstract public String name();
}

class IntNode extends TypeNode {
  public IntNode() {
  }

  public Type myType(int array, int num) { 
    if (array == -1)
      return Type.Int.addPointer(num); 
    else
      return Type.Int.addPointer(num + 1);
  }

  public String name() {
    return "int";
  }
}

class BoolNode extends TypeNode
{
  public BoolNode() {
  }

  public Type myType(int array, int num) { 
    if (array == -1)
      return Type.Bool.addPointer(num);
    else
      return Type.Bool.addPointer(num + 1);
  }

  public String name() {
    return "bool";
  }
}

class VoidNode extends TypeNode
{
  public VoidNode() {
  }

  public Type myType(int array, int num) {
    if (array == -1)
      return Type.Void.addPointer(num);
    else
      return Type.Void.addPointer(num + 1);
  }

  public String name() {
    return "void";
  }
}

class StructTypeNode extends TypeNode {
  public StructTypeNode(IdNode name) {
    myId = name;
  }

  public Type myType(int array, int num) {
    if (array == -1)
      return Type.Struct.addPointer(num);
    else
      return Type.Struct.addPointer(num + 1);
  }

  public String name() {
    return myId.getName();
  }

  private IdNode myId;
}
