/* expressions.java
 * Author: John Lenz <jelenz@wisc.edu>
 * (C) Copyright 2003 John Lenz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.util.*;
import java.io.*;

// **********************************************************************
// ExpNode and its subclasses
// **********************************************************************

abstract class ExpNode extends ASTnode {
  abstract public int getLine();
  abstract public int getChar();
  abstract public Type checkTypes();
  public void updateNames(SymTab table) {}
  public void codeGen() {}
  public void codeGen(String trueLab, String falseLab) {}
  public void codeGenStore() {}
  public void codeGenAddress() {}
  public boolean isLeftHandSize() { return false; }
}

class IntLitNode extends ExpNode {
  public IntLitNode(int lineNum, int charNum, int intVal) {
    myLineNum = lineNum;
    myCharNum = charNum;
    myIntVal = intVal;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print(myIntVal);
  }

  public Type checkTypes() {
    return Type.Int;
  }

  public void codeGen() {
    Codegen.generate("li", Codegen.T0, myIntVal);
    Codegen.genPush(Codegen.T0);
  }

  public int getLine() { return myLineNum; }
  public int getChar() { return myCharNum; }

  private int myLineNum;
  private int myCharNum;
  private int myIntVal;
}

class NullNode extends IntLitNode {
  public NullNode(int lineNum, int charNum) {
    super(lineNum, charNum, 0);
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("NULL");
  }

  public Type checkTypes() {
    return Type.NULL;
  }
}

class StringLitNode extends ExpNode {
  public StringLitNode(int lineNum, int charNum, String strVal) {
    myLineNum = lineNum;
    myCharNum = charNum;
    myStrVal = strVal;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print(myStrVal);
  }

  public Type checkTypes() {
    return Type.String;
  }

  public void codeGen() {
    String strLabel = Codegen.nextLabel();
    Codegen.generate("la", Codegen.T0, strLabel);
    Codegen.genPush(Codegen.T0);
    Codegen.generate(".data");
    Codegen.generateLabeled(strLabel, ".asciiz ", "", myStrVal);
    Codegen.generate(".text");
  }

  public int getLine() { return myLineNum; }
  public int getChar() { return myCharNum; }
  public String str() { return myStrVal; }

  private int myLineNum;
  private int myCharNum;
  private String myStrVal;
}

class TrueNode extends ExpNode {
  public TrueNode(int lineNum, int charNum) {
    myLineNum = lineNum;
    myCharNum = charNum;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("true");
  }

  public Type checkTypes() {
    return Type.Bool;
  }

  public void codeGen() {
    Codegen.generate("li", Codegen.T0, 1);
    Codegen.genPush(Codegen.T0);
  }

  public void codeGen(String trueLab, String falseLab) {
    Codegen.generate("b", trueLab);
  }

  public int getLine() { return myLineNum; }
  public int getChar() { return myCharNum; }

  private int myLineNum;
  private int myCharNum;
}

class FalseNode extends ExpNode {
  public FalseNode(int lineNum, int charNum) {
    myLineNum = lineNum;
    myCharNum = charNum;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("false");
  }

  public Type checkTypes() {
    return Type.Bool;
  }

  public void codeGen() {
    Codegen.generate("li", Codegen.T0, 0);
    Codegen.genPush(Codegen.T0);
  }

  public void codeGen(String trueLab, String falseLab) {
    Codegen.generate("b", falseLab);
  }

  public int getLine() { return myLineNum; }
  public int getChar() { return myCharNum; }

  private int myLineNum;
  private int myCharNum;
}

class SizeOfNode extends ExpNode {
  public SizeOfNode(IdNode name) {
    myId = name;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("sizeof(");
    myId.unparse(p, 0);
    p.print(")");
  }

  public Type checkTypes() {
    if (!myId.sym.type().equals(Type.Struct)) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Operand is not a structure");
      return Type.Error;
    } else {
      return Type.Int;
    }
  }

  public void updateNames(SymTab table) {
    myId.sym = table.globalLookup(myId.getName());
    if (myId.sym == null)
      Errors.fatal(myId.getLine(), myId.getChar(), "Undeclared indentifier");
  }

  public void codeGen() {
    Codegen.generate("li", Codegen.T0, myId.sym.size() * Codegen.wordSize);
    Codegen.genPush(Codegen.T0);
  }

  public int getLine() { return myId.getLine(); }
  public int getChar() { return myId.getChar(); }

  public IdNode myId;
}

class IdNode extends ExpNode {
  public IdNode(int lineNum, int charNum, String strVal) {
    myLineNum = lineNum;
    myCharNum = charNum;
    myStrVal = strVal;
    sym = null;
  }

  public void unparse(PrintWriter p, int indent) {
    if (sym != null)
      p.print(sym);
    else
      p.print(myStrVal);
  }

  /// Lookup name in symbol table and set sym accordingly.
  public void updateNames(SymTab table) {
    sym = table.globalLookup(myStrVal);
    if (sym == null)
      Errors.fatal(myLineNum, myCharNum, "Undeclared identifier");
  }

  public Type checkTypes() {
    return sym.type();
  }

  public void codeGen() {
    if (sym.offset() < 0) {
      Codegen.generate("la", Codegen.T0, "_" + sym.name());
      if (!sym.isArray())
        Codegen.generateIndexed("lw", Codegen.T0, Codegen.T0, 0);
    } else {
      if (sym.isArray()) {
        if (Codegen.Wisc && (sym.offset() < -16 || sym.offset() > 15)) {
          Codegen.generate("li", Codegen.T0, sym.offset());
          Codegen.generate("sub", Codegen.T0, Codegen.FP, Codegen.T0);
        } else {
          Codegen.generate(Codegen.SUB, Codegen.T0, Codegen.FP, sym.offset());
        }
      } else {
        if (Codegen.Wisc && (sym.offset() < -16 || sym.offset() > 15)) {
          Codegen.generate("li", Codegen.T1, -sym.offset());
          Codegen.generate("lw", Codegen.T0, Codegen.FP, Codegen.T1);
        } else {
          Codegen.generateIndexed("lw", Codegen.T0, Codegen.FP, -sym.offset());
        }
      }
    }
    Codegen.genPush(Codegen.T0);
  }

  public void codeGenStore() {
    Codegen.genPop(Codegen.T0);
    if (sym.offset() < 0) {
      Codegen.generate("la", Codegen.T1, "_" + sym.name());
      Codegen.generateIndexed("sw", Codegen.T0, Codegen.T1, 0);
    } else {
      if (Codegen.Wisc && (sym.offset() < -16 || sym.offset() > 15)) {
        Codegen.generate("li", Codegen.T1, -sym.offset());
        Codegen.generate("sw", Codegen.T0, Codegen.FP, Codegen.T1);
      } else {
        Codegen.generateIndexed("sw", Codegen.T0, Codegen.FP, -sym.offset());
      }
    }
  }

  public void codeGenAddress() {
    if (sym.offset() < 0) {
      Codegen.generate("la", Codegen.T0, "_" + sym.name());
    } else {
      if (Codegen.Wisc && (sym.offset() < -16 || sym.offset() > 15)) {
        Codegen.generate("li", Codegen.T0, sym.offset());
        Codegen.generate("sub", Codegen.T0, Codegen.FP, Codegen.T0);
      } else {
        Codegen.generate(Codegen.SUB, Codegen.T0, Codegen.FP, sym.offset());
      }
    }
    Codegen.genPush(Codegen.T0);
  }
    

  public void codeGen(String trueLab, String falseLab) {
    if (sym.offset() < 0) {
      Codegen.generate("la", Codegen.T0, "_" + sym.name());
      Codegen.generateIndexed("lw", Codegen.T0, Codegen.T0, 0);
    } else {
      if (Codegen.Wisc && (sym.offset() < -16 || sym.offset() > 15)) {
        Codegen.generate("li", Codegen.T1, -sym.offset());
        Codegen.generate("lw", Codegen.T0, Codegen.FP, Codegen.T1);
      } else {
        Codegen.generateIndexed("lw", Codegen.T0, Codegen.FP, -sym.offset());
      }
    }
    Codegen.generate("beqz", Codegen.T0, falseLab);
    Codegen.generate("b", trueLab);
  }

  public boolean isLeftHandSize() { return true; }

  public Sym sym;
  public String getName() { return myStrVal; }
  public int getLine() { return myLineNum; }
  public int getChar() { return myCharNum; }

  private int myLineNum;
  private int myCharNum;
  private String myStrVal;
}

class ArrayExpNode extends ExpNode {
  public ArrayExpNode(ExpNode lhs, ExpNode exp) {
    myLhs = lhs;
    myExp = exp;
    elementSize = -1;
    lhs = null;
  }

  public void unparse(PrintWriter p, int indent) {
    myLhs.unparse(p, indent+1);
    p.print("[");
    myExp.unparse(p, indent+1);
    p.print("]");
  }

  public void updateNames(SymTab table) {
    myLhs.updateNames(table);
    myExp.updateNames(table);
  }

  public Type checkTypes() {
    Type t = myExp.checkTypes();
    Type lhs = myLhs.checkTypes();

    if (lhs.equals(Type.Struct.addPointer(1))) {
      Sym s = cmm.symTable.globalLookup(lhs.toString());
      if (s == null) {
        System.err.println("Structure type not in symbol table:" + lhs.toString());
        System.exit(-1);
      }
      elementSize = s.size();
    }

    boolean noarray = false;
    if (lhs.numPointers() <= 0) {
      Errors.fatal(myLhs.getLine(), myLhs.getChar(), "Index applied to non-array operand");
      noarray = true;
    }
    
    if (!t.equals(Type.Int) && !t.equals(Type.Error)) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Non-int expression used as an array index");
    }
    
    if (noarray)
      return Type.Error;
    else 
      return lhs.addPointer(-1);
  }

  public void codeGen() {
    codeGenAddress();

    Codegen.genPop(Codegen.T0);
    Codegen.generateIndexed("lw", Codegen.T0, Codegen.T0, 0);
    Codegen.genPush(Codegen.T0);
  }

  public void codeGenStore() {
    codeGenAddress();

    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generateIndexed("sw", Codegen.T1, Codegen.T0, 0);
  }

  public void codeGenAddress() {
    myLhs.codeGen();
    myExp.codeGen();

    Codegen.genPop(Codegen.T1);
    if (elementSize > 1) {
      if (Codegen.Wisc) {
        Codegen.generate("move", Codegen.V0, Codegen.T1);
        Codegen.generate("li", Codegen.V1, elementSize);
        Codegen.generate("la", Codegen.T0, "multiply");
        Codegen.generate("jalr", Codegen.T0);
        Codegen.generate("move", Codegen.T1, "$6");
      } else {
        Codegen.generate("mul", Codegen.T1, Codegen.T1, elementSize);
      }
    } else if (!Codegen.Wisc) {
      Codegen.generate("sll", Codegen.T1, Codegen.T1, "2");
    }
    Codegen.genPop(Codegen.T0);
    Codegen.generate(Codegen.ADD, Codegen.T0, Codegen.T0, Codegen.T1);
    Codegen.genPush(Codegen.T0);
  }

  public void codeGen(String trueLab, String falseLab) {
    codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.generate("beqz", Codegen.T0, falseLab);
    Codegen.generate("b", trueLab);
  }

  public boolean isLeftHandSize() { return true; }

  public int getLine() { return myLhs.getLine(); }
  public int getChar() { return myLhs.getChar(); }

  // 2 kids
  private ExpNode myLhs;
  private ExpNode myExp;
  private int elementSize;
}

class CallExpNode extends ExpNode {
  public CallExpNode(IdNode name, ExpListNode elist) {
    myId = name;
    myExpList = elist;
  }

  public CallExpNode(IdNode name) {
    myId = name;
    myExpList = new ExpListNode(new LinkedList());
  }

  // ** unparse **
  public void unparse(PrintWriter p, int indent) {
    myId.unparse(p, indent+1);
    p.print("(");
    if (myExpList != null) myExpList.unparse(p, indent);
    p.print(")");
  }

  public void updateNames(SymTab table) {
    myId.updateNames(table);
    myExpList.updateNames(table);
  }

  public Type checkTypes() {
    if (! myId.sym.type().equals(Type.Func)) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Attempt to call a non-function");
      return Type.Error;
    }

    if (myId.sym.params().size() - 1 != myExpList.numExprs()) {
      Errors.fatal(myId.getLine(), myId.getChar(), "Function call with wrong number of args");
      return Type.Error;
    }

    Iterator it = myId.sym.params().iterator();
    try {
      Type ret = (Type) it.next();
      myExpList.checkTypes(it);
      return ret;
    } catch (NoSuchElementException ex) {
      System.err.println("unexpected NoSuchElementException in CallExpNode.updateNames");
      ex.printStackTrace();
      System.exit(-1);
    }
    // can't reach this statement but compiler still complains
    return Type.Error;
  }

  public void codeGen() {
    myExpList.codeGen();
    if (Codegen.Wisc) {
      if (myId.sym.name().equals("main"))
        Codegen.generate("la", Codegen.T0, "main");
      else
        Codegen.generate("la", Codegen.T0, "_" + myId.sym.name());
      Codegen.generate("jalr", Codegen.T0);
    } else {
      if (myId.sym.name().equals("main"))
        Codegen.generate("jal", "main");
      else
        Codegen.generate("jal", "_" + myId.sym.name());
    }
    if (!voidFunction())
      Codegen.genPush(Codegen.V0);
  }

  public void codeGen(String trueLab, String falseLab) {
    myExpList.codeGen();
    if (Codegen.Wisc) {
      if (myId.sym.name().equals("main"))
        Codegen.generate("la", Codegen.T0, "main");
      else
        Codegen.generate("la", Codegen.T0, "_" + myId.sym.name());
      Codegen.generate("jalr", Codegen.T0);
    } else {
      if (myId.sym.name().equals("main"))
        Codegen.generate("jal", "main");
      else
        Codegen.generate("jal", "_" + myId.sym.name());
    }
    Codegen.generate("beqz", Codegen.V0, falseLab);
    Codegen.generate("b", trueLab);
  }

  public int getLine() { return myId.getLine(); }
  public int getChar() { return myId.getChar(); }
  public boolean voidFunction() { return ((Type)myId.sym.params().getFirst()).equals(Type.Void); }

  // 2 kids
  private IdNode myId;
  private ExpListNode myExpList;  // possibly null
}

class StructRefNode extends ExpNode {
  public StructRefNode(ExpNode exp, IdNode member) {
    myExp = exp;
    myMember = member;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp.unparse(p, 0);
    p.print(".");
    myMember.unparse(p, 0);
    p.print(")");
  }

  public void updateNames(SymTab table) {
    myExp.updateNames(table);
  }

  public Type checkTypes() {
    Type expType = myExp.checkTypes();
    if (!expType.equals(Type.Error)) {
      if (expType.equals(Type.Struct) && expType.numPointers() <= 0) {
        structSym = cmm.symTable.globalLookup(expType.toString());
        if (structSym == null) {
          System.err.println("Structure name not found in symbol table:" + expType.toString());
          System.exit(-1);
        }
        myMember.sym = cmm.symTable.globalLookup(structSym.name() + "." + myMember.getName());
        if (myMember.sym == null) {
          Errors.fatal(myMember.getLine(), myMember.getChar(), "Invalid member name for structure");
        }
      } else {
        Errors.fatal(myExp.getLine(), myExp.getChar(), "Non-struct used in structure access");
      }
    }
    if (myMember.sym == null)
      return Type.Error;
    else
      return myMember.sym.type();
  }

  public void codeGen() {
    codeGenAddress();

    if (!myMember.sym.isArray()) {
      Codegen.genPop(Codegen.T0);
      Codegen.generateIndexed("lw", Codegen.T0, Codegen.T0, 0);
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGenStore() {
    codeGenAddress();

    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generateIndexed("sw", Codegen.T1, Codegen.T0, 0);
  }

  public void codeGenAddress() {
    myExp.codeGenAddress();
    if (myMember.sym.offset() != 0) {
      Codegen.genPop(Codegen.T0);
      if (Codegen.Wisc && (myMember.sym.offset() < -16 || myMember.sym.offset() > 15)) {
        Codegen.generate("li", Codegen.T1, myMember.sym.offset());
        Codegen.generate("add", Codegen.T0, Codegen.T0, Codegen.T1);
      } else {
        Codegen.generate(Codegen.ADD, Codegen.T0, Codegen.T0, myMember.sym.offset());
      }
      Codegen.genPush(Codegen.T0);
    }
  }

  public boolean isLeftHandSize() { return true; }

  public void codeGen(String trueLab, String falseLab) {
    codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.generate("beqz", Codegen.T0, falseLab);
    Codegen.generate("b", trueLab);
  }

  public int getLine() { return myExp.getLine(); }
  public int getChar() { return myExp.getChar(); }

  // 2 kids
  private ExpNode myExp;
  private IdNode myMember;
  private Sym structSym;
}

abstract class UnaryExpNode extends ExpNode {
  public UnaryExpNode(ExpNode exp) {
    myExp = exp;
  }

  public void updateNames(SymTab table) {
    myExp.updateNames(table);
  }

  public Type checkTypes() {
    Type t = myExp.checkTypes();
    if (t.equals(Type.Error)) return Type.Error;

    if (! t.equals(validType)) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), errorStr);
      return Type.Error;
    }
    return validType;
  }

  public int getLine() { return myExp.getLine(); }
  public int getChar() { return myExp.getChar(); }

  // one child
  protected ExpNode myExp;
  protected String errorStr;
  protected Type validType;
}

abstract class BinaryExpNode extends ExpNode {
  public BinaryExpNode(ExpNode exp1, ExpNode exp2) {
    myExp1 = exp1;
    myExp2 = exp2;
  }

  public void updateNames(SymTab table) {
    myExp1.updateNames(table);
    myExp2.updateNames(table);
  }

  public Type checkTypes() {
    Type t1 = myExp1.checkTypes();
    Type t2 = myExp2.checkTypes();
    boolean error = false;

    if (t1.equals(Type.Error) || t2.equals(Type.Error)) error = true;

    if ((!t1.equals(validType) && !t1.equals(Type.Error)) || t1.numPointers() > 0) {
      Errors.fatal(myExp1.getLine(), myExp1.getChar(), errorStr);
      error = true;
    }
    if ((!t2.equals(validType) && !t2.equals(Type.Error)) || t2.numPointers() > 0) {
      Errors.fatal(myExp2.getLine(), myExp2.getChar(), errorStr);
      error = true;
    }
    if (error)
      return Type.Error;
    else
      return expType;
  }

  public int getLine() { return myExp1.getLine(); }
  public int getChar() { return myExp1.getChar(); }

  // two kids
  protected ExpNode myExp1;
  protected ExpNode myExp2;
  protected String errorStr;
  protected Type validType;
  protected Type expType;
}


// **********************************************************************
// Subclasses of UnaryExpNode
// **********************************************************************

class UnaryMinusNode extends UnaryExpNode {
  public UnaryMinusNode(ExpNode exp) {
    super(exp);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(-");
    myExp.unparse(p, indent+1);
    p.print(")");
  }

  public void codeGen() {
    myExp.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.generate("neg", Codegen.T0, Codegen.T0);
    Codegen.genPush(Codegen.T0);
  }
}

class NotNode extends UnaryExpNode {
  public NotNode(ExpNode exp) {
    super(exp);
    errorStr = "Logical operator applied to non-bool operand";
    validType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(!");
    myExp.unparse(p, indent+1);
    p.print(")");
  }

  public void codeGen() {
    myExp.codeGen();
    Codegen.genPop(Codegen.T0);
    if (Codegen.Wisc) {
      Codegen.generate("li", Codegen.T1, "1");
      String l = Codegen.nextLabel();
      Codegen.generate("beqz", Codegen.T0, l);
      Codegen.generate("li", Codegen.T1, "0");
      Codegen.genLabel(l);
      Codegen.genPush(Codegen.T1);
    } else {
      Codegen.generate("seq", Codegen.T0, Codegen.T0, "0");
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    Codegen.genPop(Codegen.T0);
    Codegen.generate("beqz", Codegen.T0, falseLab);
    Codegen.generate("b", trueLab);
  }
}

class AddrOfNode extends UnaryExpNode {
  public AddrOfNode(ExpNode exp) {
    super(exp);
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(&");
    myExp.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    if (myExp.isLeftHandSize())
      return myExp.checkTypes().addPointer(1);
    else {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Address-of operator applied to expression");
      return Type.Error;
    }
  }

  public void codeGen() {
    myExp.codeGenAddress();
  }
}

class DeRefNode extends UnaryExpNode {
  public DeRefNode(ExpNode exp) {
    super(exp);
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(*");
    myExp.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    Type t = myExp.checkTypes();
    if (t.numPointers() == 0) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Dereference applied to base type");
      return Type.Error;
    } else
      return myExp.checkTypes().addPointer(-1);
  }

  public void codeGen() {
    myExp.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.generateIndexed("lw", Codegen.T0, Codegen.T0, 0);
    Codegen.genPush(Codegen.T0);
  }

  public void codeGenStore() {
    myExp.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generateIndexed("sw", Codegen.T1, Codegen.T0, 0);
  }

  public void codeGen(String trueLab, String falseLab) {
    codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.generate("beqz", Codegen.T0, falseLab);
    Codegen.generate("b", trueLab);
  }

  public void codeGenAddress() {
    myExp.codeGen();
  }

  public boolean isLeftHandSize() { return true; }
}

// **********************************************************************
// Subclasses of BinaryExpNode
// **********************************************************************

class PlusNode extends BinaryExpNode {
  public PlusNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Int;
    pointer = 0;
    elementSize = -1;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" + ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    Type t1 = myExp1.checkTypes();
    Type t2 = myExp2.checkTypes();
    boolean error = false;

    if (t1.equals(Type.Error) || t2.equals(Type.Error)) error = true;

    if (t1.equals(Type.Int)) {
      if (t2.numPointers() > 0) {
        t1 = t2;
        pointer = 2;
      } else if (!t2.equals(Type.Int) && !t2.equals(Type.Error)) {
        Errors.fatal(myExp2.getLine(), myExp2.getChar(), errorStr);
        error = true;
      }
    } else if (t1.numPointers() > 0) {
      if (!t2.equals(Type.Int) && !t2.equals(Type.Error)) {
        Errors.fatal(myExp2.getLine(), myExp2.getChar(), errorStr);
        error = true;
      } else {
        pointer = 1;
      }
    } else if (!t1.equals(Type.Error)) {
      Errors.fatal(myExp1.getLine(), myExp1.getChar(), errorStr);
      error = true;
    }

    if (t1.equals(Type.Struct.addPointer(1))) {
      Sym s = cmm.symTable.globalLookup(t1.toString());
      if (s == null) {
        System.err.println("Structure type not in symbol table:" + t1.toString());
        System.exit(-1);
      }
      elementSize = s.size();
    }
       
    if (error)
      return Type.Error;
    else if (pointer != 0)
      return t1;
    else
      return Type.Int;
  }

  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);

    String reg = "";
    String otherReg = "";
    if (pointer == 1) {
      reg = Codegen.T0;
      otherReg = Codegen.T1;
    } else if (pointer == 2) {
      reg = Codegen.T1;
      otherReg = Codegen.T0;
    }
    
    if (pointer != 0) {
      if (elementSize > 1) {
        if (Codegen.Wisc) {
          Codegen.genPush(otherReg);
          Codegen.generate("move", Codegen.V0, reg);
          Codegen.generate("li", Codegen.V1, elementSize);
          Codegen.generate("la", reg, "multiply");
          Codegen.generate("jalr", reg);
          Codegen.generate("move", reg, "$6");
          Codegen.genPop(otherReg);
        } else {
          Codegen.generate("mul", reg, reg, elementSize);
        }
      } else if (!Codegen.Wisc) {
        Codegen.generate("sll", reg, reg, "2");
      }
    }
    Codegen.generate("add", Codegen.T0, Codegen.T0, Codegen.T1);
    Codegen.genPush(Codegen.T0);
  }

  private int pointer;
  private boolean array;
  private Type ptrType;
  private int elementSize;
}

class MinusNode extends BinaryExpNode {
  public MinusNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Int;
    pointer = 0;
    elementSize = -1;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" - ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    Type t1 = myExp1.checkTypes();
    Type t2 = myExp2.checkTypes();
    boolean error = false;

    if (t1.equals(Type.Error) || t2.equals(Type.Error)) error = true;

    if (t1.equals(Type.Int)) {
      if (!t2.equals(Type.Int) && !t2.equals(Type.Error)) {
        Errors.fatal(myExp2.getLine(), myExp2.getChar(), errorStr);
        error = true;
      }
    } else if (t1.numPointers() > 0) {
      if (t2.numPointers() > 0) {
        pointer = 3;
      } else {
        if (!t2.equals(Type.Int) && !t2.equals(Type.Error)) {
          Errors.fatal(myExp2.getLine(), myExp2.getChar(), errorStr);
          error = true;
        } else {
          pointer = 1;
        }
      }
    } else if (!t1.equals(Type.Error)) {
      Errors.fatal(myExp1.getLine(), myExp1.getChar(), errorStr);
      error = true;
    }

    if (t1.equals(Type.Struct.addPointer(1))) {
      Sym s = cmm.symTable.globalLookup(t1.toString());
      if (s == null) {
        System.err.println("Structure type not in symbol table:" + t1.toString());
        System.exit(-1);
      }
      elementSize = s.size();
    }

    if (error)
      return Type.Error;
    else if (pointer == 1)
      return t1;
    else /*if (pointer == 3) and else */
      return Type.Int;
  }

  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    
    if (pointer == 1) {
      if (elementSize > 1) {
        if (Codegen.Wisc) {
          Codegen.genPush(Codegen.T1);
          Codegen.generate("move", Codegen.V0, Codegen.T0);
          Codegen.generate("li", Codegen.V1, elementSize);
          Codegen.generate("la", Codegen.T0, "multiply");
          Codegen.generate("jalr", Codegen.T0);
          Codegen.generate("move", Codegen.T0, "$6");
          Codegen.genPop(Codegen.T1);
        } else {
          Codegen.generate("mul", Codegen.T0, Codegen.T0, elementSize);
        }
      } else if (!Codegen.Wisc) {
        Codegen.generate("sll", Codegen.T0, Codegen.T0, "2");
      }
    }
    Codegen.generate("sub", Codegen.T0, Codegen.T1, Codegen.T0);
    if (pointer == 3) {
      if (elementSize > 1) {
        if (Codegen.Wisc) {
          Codegen.generate("move", Codegen.V0, Codegen.T0);
          Codegen.generate("li", Codegen.V1, elementSize);
          Codegen.generate("la", Codegen.T0, "divide");
          Codegen.generate("jalr", Codegen.T0);
          Codegen.generate("move", Codegen.T0, "$6");
        } else {
          Codegen.generate("div", Codegen.T0, Codegen.T0, elementSize);
        }
      } else if (!Codegen.Wisc)
        Codegen.generate("sra", Codegen.T0, Codegen.T0, "2");
    }
    Codegen.genPush(Codegen.T0);
  }

  private int pointer;
  private int elementSize;
}

class TimesNode extends BinaryExpNode {
  public TimesNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Int;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" * ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    if (Codegen.Wisc) {
      Codegen.genPop(Codegen.V0);
      Codegen.genPop(Codegen.V1);
      Codegen.generate("la", Codegen.T0, "multiply");
      Codegen.generate("jalr", Codegen.T0);
      Codegen.genPush("$6");
    } else {
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      Codegen.generate("mul", Codegen.T0, Codegen.T0, Codegen.T1);
      Codegen.genPush(Codegen.T0);
    }
  }
}

class DivideNode extends BinaryExpNode {
  public DivideNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Int;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" / ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    if (Codegen.Wisc) {
      Codegen.genPop(Codegen.V1);
      Codegen.genPop(Codegen.V0);
      Codegen.generate("la", Codegen.T0, "divide");
      Codegen.generate("jalr", Codegen.T0);
      Codegen.genPush("$6");
    } else {
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      Codegen.generate("div", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }
}

class ModuloNode extends BinaryExpNode {
  public ModuloNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Arithmetic operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Int;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" % ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    if (Codegen.Wisc) {
      Codegen.genPop(Codegen.V1);
      Codegen.genPop(Codegen.V0);
      Codegen.generate("la", Codegen.T0, "divide");
      Codegen.generate("jalr", Codegen.T0);
      Codegen.genPush(Codegen.T0);
    } else {
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      Codegen.generate("rem", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }
}

class AndNode extends BinaryExpNode {
  public AndNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Logical operator applied to non-bool operand";
    validType = Type.Bool;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" && ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public void codeGen() {
    myExp1.codeGen();
    Codegen.generateIndexed("lw", Codegen.T0, Codegen.SP, Codegen.wordSize);
    String lab = Codegen.nextLabel();
    Codegen.generate("beqz", Codegen.T0, lab);
    Codegen.generate(Codegen.ADD, Codegen.SP, Codegen.SP, Codegen.wordSize);
    myExp2.codeGen();
    Codegen.genLabel(lab);
  }

  public void codeGen(String trueLab, String falseLab) {
    String newLab = Codegen.nextLabel();
    myExp1.codeGen(newLab, falseLab);
    Codegen.genLabel(newLab);
    myExp2.codeGen(trueLab, falseLab);
  }
}

class OrNode extends BinaryExpNode {
  public OrNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Logical operator applied to non-bool operand";
    validType = Type.Bool;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" || ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public void codeGen() {
    myExp1.codeGen();
    Codegen.generateIndexed("lw", Codegen.T0, Codegen.SP, Codegen.wordSize);
    String lab = Codegen.nextLabel();
    Codegen.generate("bnez", Codegen.T0, lab);
    Codegen.generate(Codegen.ADD, Codegen.SP, Codegen.SP, Codegen.wordSize);
    myExp2.codeGen();
    Codegen.genLabel(lab);
  }

  public void codeGen(String trueLab, String falseLab) {
    String newLab = Codegen.nextLabel();
    myExp1.codeGen(trueLab, newLab);
    Codegen.genLabel(newLab);
    myExp2.codeGen(trueLab, falseLab);
  }
}

class EqualsNode extends BinaryExpNode {
  public EqualsNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" == ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    Type t1 = myExp1.checkTypes();
    Type t2 = myExp2.checkTypes();

    if (t1.equals(Type.Error) || t2.equals(Type.Error)) return Type.Error;

    stringCompare = false;
    if (t1.equals(Type.String)) stringCompare = true;

    if ((t1.numPointers() == 0 && t2.equals(Type.NULL)) ||
        (t2.numPointers() == 0 && t1.equals(Type.NULL))) {
      Errors.fatal(myExp1.getLine(), myExp1.getChar(), "NULL value compared to literal value");
      return Type.Error;
    } else if (!t1.equals(Type.NULL) && !t2.equals(Type.NULL)) {
      if (! t1.equals(t2)) {
        Errors.fatal(myExp1.getLine(), myExp1.getChar(), "Type mismatch");
        return Type.Error;
      } else if (t1.equals(Type.Func)) {
        Errors.fatal(myExp1.getLine(), myExp1.getChar(), "Equality operator applied to functions");
        return Type.Error;
      }
    }
    return Type.Bool;
  }

  public void codeGen() {
    if (stringCompare) {
      if (((StringLitNode)myExp1).str().equals(((StringLitNode)myExp2).str())) {
        Codegen.generate("li", Codegen.T0, "1");
      } else {
        Codegen.generate("li", Codegen.T0, "0");
      }
      Codegen.genPush(Codegen.T0);
    } else {
      myExp1.codeGen();
      myExp2.codeGen();
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      if (Codegen.Wisc) {
        Codegen.generate("li", Codegen.V0, "1");
        String lab = Codegen.nextLabel();
        Codegen.generate("beq", Codegen.T0, Codegen.T1, lab);
        Codegen.generate("li", Codegen.V0, "0");
        Codegen.genLabel(lab);
        Codegen.genPush(Codegen.V0);
      } else {
        Codegen.generate("seq", Codegen.T0, Codegen.T0, Codegen.T1);
        Codegen.genPush(Codegen.T0);
      }
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    if (stringCompare) {
      if (((StringLitNode)myExp1).str().equals(((StringLitNode)myExp2).str())) {
        Codegen.generate("b", trueLab);
      } else {
        Codegen.generate("b", falseLab);
      }
    } else {
      myExp1.codeGen();
      myExp2.codeGen();
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      Codegen.generate("beq", Codegen.T0, Codegen.T1, trueLab);
      Codegen.generate("b", falseLab);
    }
  }

  private boolean stringCompare;
}

class NotEqualsNode extends BinaryExpNode {
  public NotEqualsNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" != ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public Type checkTypes() {
    Type t1 = myExp1.checkTypes();
    Type t2 = myExp2.checkTypes();

    if (t1.equals(Type.Error) || t2.equals(Type.Error)) return Type.Error;

    stringCompare = false;
    if (t1.equals(Type.String)) stringCompare = true;

    if ((t1.numPointers() == 0 && t2.equals(Type.NULL)) ||
        (t2.numPointers() == 0 && t1.equals(Type.NULL))) {
      Errors.fatal(myExp1.getLine(), myExp1.getChar(), "NULL value compared to literal value");
      return Type.Error;
    } else if (!t1.equals(Type.NULL) && !t2.equals(Type.NULL)) {
      if (! t1.equals(t2)) {
        Errors.fatal(myExp1.getLine(), myExp1.getChar(), "Type mismatch");
        return Type.Error;
      } else if (t1.equals(Type.Func)) {
        Errors.fatal(myExp1.getLine(), myExp1.getChar(), "Equality operator applied to functions");
        return Type.Error;
      }
    }
    return Type.Bool;
  }

  public void codeGen() {
    if (stringCompare) {
      if (((StringLitNode)myExp1).str().equals(((StringLitNode)myExp2).str())) {
        Codegen.generate("li", Codegen.T0, "0");
      } else {
        Codegen.generate("li", Codegen.T0, "1");
      }
      Codegen.genPush(Codegen.T0);
    } else {
      myExp1.codeGen();
      myExp2.codeGen();
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      if (Codegen.Wisc) {
        Codegen.generate("li", Codegen.V0, "1");
        String lab = Codegen.nextLabel();
        Codegen.generate("bne", Codegen.T0, Codegen.T1, lab);
        Codegen.generate("li", Codegen.V0, "0");
        Codegen.genLabel(lab);
        Codegen.genPush(Codegen.V0);
      } else {
        Codegen.generate("sne", Codegen.T0, Codegen.T0, Codegen.T1);
        Codegen.genPush(Codegen.T0);
      }
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    if (stringCompare) {
      if (((StringLitNode)myExp1).str().equals(((StringLitNode)myExp2).str())) {
        Codegen.generate("b", falseLab);
      } else {
        Codegen.generate("b", trueLab);
      }
    } else {
      myExp1.codeGen();
      myExp2.codeGen();
      Codegen.genPop(Codegen.T0);
      Codegen.genPop(Codegen.T1);
      Codegen.generate("bne", Codegen.T0, Codegen.T1, trueLab);
      Codegen.generate("b", falseLab);
    }
  }

  private boolean stringCompare;
}

class LessNode extends BinaryExpNode {
  public LessNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Relational operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" < ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    if (Codegen.Wisc) {
      Codegen.generate("li", Codegen.V0, "1");
      String lab = Codegen.nextLabel();
      Codegen.generate("blt", Codegen.T1, Codegen.T0, lab);
      Codegen.generate("li", Codegen.V0, "0");
      Codegen.genLabel(lab);
      Codegen.genPush(Codegen.V0);
    } else {
      Codegen.generate("slt", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generate("blt", Codegen.T1, Codegen.T0, trueLab);
    Codegen.generate("b", falseLab);
  }
}

class GreaterNode extends BinaryExpNode {
  public GreaterNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Relational operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" > ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }

  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    if (Codegen.Wisc) {
      Codegen.generate("li", Codegen.V0, "1");
      String lab = Codegen.nextLabel();
      Codegen.generate("bgt", Codegen.T1, Codegen.T0, lab);
      Codegen.generate("li", Codegen.V0, "0");
      Codegen.genLabel(lab);
      Codegen.genPush(Codegen.V0);
    } else {
      Codegen.generate("sgt", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generate("bgt", Codegen.T1, Codegen.T0, trueLab);
    Codegen.generate("b", falseLab);
  }
}

class LessEqNode extends BinaryExpNode {
  public LessEqNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Relational operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" <= ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    if (Codegen.Wisc) {
      Codegen.generate("li", Codegen.V0, "1");
      String lab = Codegen.nextLabel();
      Codegen.generate("ble", Codegen.T1, Codegen.T0, lab);
      Codegen.generate("li", Codegen.V0, "0");
      Codegen.genLabel(lab);
      Codegen.genPush(Codegen.V0);
    } else {
      Codegen.generate("sle", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generate("ble", Codegen.T1, Codegen.T0, trueLab);
    Codegen.generate("b", falseLab);
  }
}

class GreaterEqNode extends BinaryExpNode {
  public GreaterEqNode(ExpNode exp1, ExpNode exp2) {
    super(exp1, exp2);
    errorStr = "Relational operator applied to non-numeric operand";
    validType = Type.Int;
    expType = Type.Bool;
  }

  public void unparse(PrintWriter p, int indent) {
    p.print("(");
    myExp1.unparse(p, indent+1);
    p.print(" >= ");
    myExp2.unparse(p, indent+1);
    p.print(")");
  }
  public void codeGen() {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    if (Codegen.Wisc) {
      Codegen.generate("li", Codegen.V0, "1");
      String lab = Codegen.nextLabel();
      Codegen.generate("bge", Codegen.T1, Codegen.T0, lab);
      Codegen.generate("li", Codegen.V0, "0");
      Codegen.genLabel(lab);
      Codegen.genPush(Codegen.V0);
    } else {
      Codegen.generate("sge", Codegen.T0, Codegen.T1, Codegen.T0);
      Codegen.genPush(Codegen.T0);
    }
  }

  public void codeGen(String trueLab, String falseLab) {
    myExp1.codeGen();
    myExp2.codeGen();
    Codegen.genPop(Codegen.T0);
    Codegen.genPop(Codegen.T1);
    Codegen.generate("bge", Codegen.T1, Codegen.T0, trueLab);
    Codegen.generate("b", falseLab);
  }
}
