/* statements.java
 * Author: John Lenz <jelenz@wisc.edu>
 * (C) Copyright 2003 John Lenz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import java.io.*;
import java.util.*;

// **********************************************************************
// StmtNode and its subclasses
// **********************************************************************

abstract class StmtNode extends ASTnode {
  /// The type is what the return type of the enclosing function.
  abstract public void checkTypes(Type ret);
  abstract public int updateNames(SymTab table, int off);
  abstract public void codeGen(String retLabel);
}

class ReadStmtNode extends StmtNode {
  public ReadStmtNode(ExpNode e) {
    myExp = e;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("cin >> ");
    myExp.unparse(p, 0);
    p.println(";");
  }

  public void checkTypes(Type ret) {
    Type t = myExp.checkTypes();
    if (t.equals(Type.Func)) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Attempt to read a function");
    } else if (t.numPointers() > 0) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Attempt to read a pointer");
    }
  }

  public int updateNames(SymTab table, int off) {
    myExp.updateNames(table);
    return off;
  }

  public void codeGen(String retLabel) {
    if (Codegen.Wisc) {
    } else {
      Codegen.generateWithComment("li", "READ", Codegen.V0, "5");
      Codegen.generate("syscall");
      Codegen.genPush(Codegen.V0);
      myExp.codeGenStore();
    } 
  }

  // 1 kid (actually can only be an IdNode or an ArrayExpNode)
  private ExpNode myExp;
}

class WriteStmtNode extends StmtNode {
  public WriteStmtNode(ExpNode exp) {
    myExp = exp;
  }

  // ** unparse **
  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("cout << ");
    myExp.unparse(p, 0);
    p.println(";");
  }

  public int updateNames(SymTab table, int off) {
    myExp.updateNames(table);
    return off;
  }

  public void checkTypes(Type ret) {
    outType = myExp.checkTypes();
    if (outType.equals(Type.Func)) {
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Attempt to write a function");
    } else if (outType.numPointers() > 0) {
      Errors.warn(myExp.getLine(), myExp.getChar(), "Attempt to write a pointer");
    }
  }

  public void codeGen(String retLabel) {
    if (Codegen.Wisc) {
      myExp.codeGen();
      Codegen.genPop(Codegen.T0);
      if (outType.equals(Type.String)) {
        String loop = Codegen.nextLabel();
        String end = Codegen.nextLabel();
        Codegen.genLabel(loop);
        Codegen.generateIndexed("lw", Codegen.T1, Codegen.T0, 0);
        Codegen.generate("beqz", Codegen.T1, end);
        Codegen.generate("wrch", Codegen.T1);
        Codegen.generate("add", Codegen.T0, Codegen.T0, 1);
        Codegen.generate("b", loop);
        Codegen.genLabel(end);
      } else {
        Codegen.generate("wrint", Codegen.T0);
      }
    } else {
      myExp.codeGen();
      Codegen.genPop(Codegen.A0);
      if (outType.equals(Type.String))
        Codegen.generateWithComment("li", "STRING", Codegen.V0, "4");
      else
        Codegen.generateWithComment("li", "WRITE", Codegen.V0, "1");
      Codegen.generate("syscall");
    }
  }

  // 1 kid
  private ExpNode myExp;
  private Type outType;
}

class AssignStmtNode extends StmtNode {
  public AssignStmtNode(ExpNode lhs, ExpNode exp) {
    myLhs = lhs;
    myExp = exp;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    myLhs.unparse(p, 0);
    p.print(" = ");
    myExp.unparse(p, 0);
    p.println(";");
  }

  public int updateNames(SymTab table, int off) {
    myLhs.updateNames(table);
    myExp.updateNames(table);
    return off;
  }

  public void checkTypes(Type ret) {
    Type lhs = myLhs.checkTypes();
    Type rhs = myExp.checkTypes();

    if (lhs.equals(Type.Error) || rhs.equals(Type.Error))
      return;

    if (!rhs.equals(Type.NULL) || lhs.numPointers() <= 0) {
      if (!lhs.equals(rhs) && (lhs.numPointers() <= 0 || !rhs.equals(Type.Void.addPointer(1)))) {
        Errors.fatal(myLhs.getLine(), myLhs.getChar(), "Type mismatch");
      } else if (lhs.equals(Type.Func)) {
        Errors.fatal(myLhs.getLine(), myLhs.getChar(), "Function assignment");
      } 
    }
  }
  
  public void codeGen(String retLabel) {
    myExp.codeGen();
    myLhs.codeGenStore();
  }

  // 2 kids
  private ExpNode myLhs;
  private ExpNode myExp;
}

class IfStmtNode extends StmtNode {
  public IfStmtNode(ExpNode exp, DeclListNode dlist, StmtListNode slist) {
    myDeclList = dlist;
    myExp = exp;
    myStmtList = slist;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("if (");
    myExp.unparse(p, 0);
    p.println(") {");
    myDeclList.unparse(p, indent+2);
    myStmtList.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
  }
  
  public int updateNames(SymTab table, int off) {
    myExp.updateNames(table);
    table.addHashtab();
      off = myDeclList.updateNames(table, off);
      off = myStmtList.updateNames(table, off);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in IfStmtNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void checkTypes(Type ret) {
    Type test = myExp.checkTypes();
    if (! test.equals(Type.Error) && ! test.equals(Type.Bool))
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Non-bool expression used as an if condition");

    myStmtList.checkTypes(ret);
  }

  public void codeGen(String retLabel) {
    Codegen.generateComment("IF THEN");
    String trueLab = Codegen.nextLabel();
    String falseLab = Codegen.nextLabel();
    myExp.codeGen(trueLab, falseLab);
    Codegen.genLabel(trueLab);
    myStmtList.codeGen(retLabel);
    Codegen.genLabel(falseLab);
  }

  // e kids
  private ExpNode myExp;
  private DeclListNode myDeclList;
  private StmtListNode myStmtList;
}

class IfElseStmtNode extends StmtNode {
  public IfElseStmtNode(ExpNode exp, DeclListNode dlist1,
                  StmtListNode slist1, DeclListNode dlist2,
                  StmtListNode slist2) {
    myExp = exp;
    myThenDeclList = dlist1;
    myThenStmtList = slist1;
    myElseDeclList = dlist2;
    myElseStmtList = slist2;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("if (");
    myExp.unparse(p, 0);
    p.println(") {");
    myThenDeclList.unparse(p, indent+2);
    myThenStmtList.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
    doIndent(p, indent);
    p.println("else {");
    myElseDeclList.unparse(p, indent+2);
    myElseStmtList.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
  }

  public int updateNames(SymTab table, int off) {
    myExp.updateNames(table);
    table.addHashtab();
      off = myThenDeclList.updateNames(table, off);
      off = myThenStmtList.updateNames(table, off);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in IfElseStmtNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    table.addHashtab();
      off = myElseDeclList.updateNames(table, off);
      off = myElseStmtList.updateNames(table, off);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in IfElseStmtNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void checkTypes(Type ret) {
    Type test = myExp.checkTypes();
    if (! test.equals(Type.Error) && ! test.equals(Type.Bool))
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Non-bool expression used as an if condition");

    myThenStmtList.checkTypes(ret);
    myElseStmtList.checkTypes(ret);
  }

  public void codeGen(String retLabel) {
    Codegen.generateComment("IF THEN ELSE");
    String trueLab = Codegen.nextLabel();
    String falseLab = Codegen.nextLabel();
    String endLab = Codegen.nextLabel();
    myExp.codeGen(trueLab, falseLab);
    Codegen.genLabel(trueLab);
    myThenStmtList.codeGen(retLabel);
    Codegen.generate("b", endLab);
    Codegen.genLabel(falseLab);
    myElseStmtList.codeGen(retLabel);
    Codegen.genLabel(endLab);
  }

  // 5 kids
  private ExpNode myExp;
  private DeclListNode myThenDeclList;
  private StmtListNode myThenStmtList;
  private StmtListNode myElseStmtList;
  private DeclListNode myElseDeclList;
}

class WhileStmtNode extends StmtNode {
  public WhileStmtNode(ExpNode exp, DeclListNode dlist, StmtListNode slist) {
    myExp = exp;
    myDeclList = dlist;
    myStmtList = slist;
  }
  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("while (");
    myExp.unparse(p, 0);
    p.println(") {");
    myDeclList.unparse(p, indent+2);
    myStmtList.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
  }

  public int updateNames(SymTab table, int off) {
    myExp.updateNames(table);
    table.addHashtab();
      off = myDeclList.updateNames(table, off);
      off = myStmtList.updateNames(table, off);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in WhileStmtNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void checkTypes(Type ret) {
    Type test = myExp.checkTypes();
    if (! test.equals(Type.Error) && ! test.equals(Type.Bool))
      Errors.fatal(myExp.getLine(), myExp.getChar(), "Non-bool expression used as a while condition");

    myStmtList.checkTypes(ret);
  }

  public void codeGen(String retLabel) {
    Codegen.generateComment("WHILE");
    String startLab = Codegen.nextLabel();
    String trueLab = Codegen.nextLabel();
    String endLab = Codegen.nextLabel();
    Codegen.genLabel(startLab);
    myExp.codeGen(trueLab, endLab);
    Codegen.genLabel(trueLab);
    myStmtList.codeGen(retLabel);
    Codegen.generate("b", startLab);
    Codegen.genLabel(endLab);
  }

  // 3 kids
  private ExpNode myExp;
  private DeclListNode myDeclList;
  private StmtListNode myStmtList;
}

class ForStmtNode extends StmtNode {
  public ForStmtNode(StmtNode init, ExpNode cond, StmtNode incr, DeclListNode dlist, StmtListNode slist) {
    myInit = init;
    myCond = cond;
    myIncr = incr;
    myDeclList = dlist;
    myStmtList = slist;
  }
  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    p.print("for (");
    myInit.unparse(p, 0);
    p.print("; ");
    myCond.unparse(p, 0);
    p.print("; ");
    myIncr.unparse(p, 0);
    p.println(") {");
    myDeclList.unparse(p, indent+2);
    myStmtList.unparse(p, indent+2);
    doIndent(p, indent);
    p.println("}");
  }

  public int updateNames(SymTab table, int off) {
    myInit.updateNames(table, off);
    myCond.updateNames(table);
    myIncr.updateNames(table, off);
    table.addHashtab();
      off = myDeclList.updateNames(table, off);
      off = myStmtList.updateNames(table, off);
    try {
      table.removeHashtab();
    } catch (EmptySymTabException e) {
      System.err.println("unexpected EmptySymTabException in WhileStmtNode.updateNames");
      e.printStackTrace();
      System.exit(-1);
    }
    return off;
  }

  public void checkTypes(Type ret) {
    myInit.checkTypes(ret);
    Type test = myCond.checkTypes();
    if (! test.equals(Type.Error) && ! test.equals(Type.Bool))
      Errors.fatal(myCond.getLine(), myCond.getChar(), "Non-bool expression used as a for condition");
    myIncr.checkTypes(ret);

    myStmtList.checkTypes(ret);
  }

  public void codeGen(String retLabel) {
    Codegen.generateComment("FOR");
    String startLab = Codegen.nextLabel();
    String trueLab = Codegen.nextLabel();
    String endLab = Codegen.nextLabel();
    myInit.codeGen(retLabel);
    Codegen.genLabel(startLab);
    myCond.codeGen(trueLab, endLab);
    Codegen.genLabel(trueLab);
    myStmtList.codeGen(retLabel);
    myIncr.codeGen(retLabel);
    Codegen.generate("b", startLab);
    Codegen.genLabel(endLab);
  }

  // 3 kids
  private StmtNode myInit;
  private ExpNode myCond;
  private StmtNode myIncr;
  private DeclListNode myDeclList;
  private StmtListNode myStmtList;
}

class CallStmtNode extends StmtNode {
  public CallStmtNode(CallExpNode call) {
    myCall = call;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    myCall.unparse(p, 0);
    p.println(";");
  }

  public int updateNames(SymTab table, int off) {
    myCall.updateNames(table);
    return off;
  }

  public void checkTypes(Type ret) {
    myCall.checkTypes();
  }

  public void codeGen(String retLabel) {
    myCall.codeGen();
    if (!myCall.voidFunction())
      Codegen.generate(Codegen.ADD, Codegen.SP, Codegen.SP, Codegen.wordSize);
  }

  // 1 kid
  private CallExpNode myCall;
}

class ReturnStmtNode extends StmtNode {
  public ReturnStmtNode(ExpNode exp) {
    myExp = exp;
  }

  public void unparse(PrintWriter p, int indent) {
    doIndent(p, indent);
    if (myExp == null) {
      p.println("return;");
    } else {
      p.print("return ");
      myExp.unparse(p, 0);
      p.println(";");
    }
  }

  public int updateNames(SymTab table, int off) {
    if (myExp != null) myExp.updateNames(table);
    return off;
  }

  public void checkTypes(Type ret) {
    if (myExp != null) {
      Type t = myExp.checkTypes();
      if (ret.equals(Type.Void) && ! t.equals(Type.Void)) 
        Errors.fatal(myExp.getLine(), myExp.getChar(), "Return with a value in a void function");
      else if (! t.equals(Type.Error) && ! t.equals(ret)) {
        Errors.fatal(myExp.getLine(), myExp.getChar(), "Bad return value");
      }
    } else if (! ret.equals(Type.Void)) {
      Errors.fatal(0, 0, "Missing return value");
    }
  }

  public void codeGen(String retLabel) {
    if (myExp != null) {
      myExp.codeGen();
      Codegen.genPop(Codegen.V0);
    }
    Codegen.generateWithComment("b", "RETURN", retLabel);
  }

  // 1 kid
  private ExpNode myExp; // possibly null
}
